-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Expo")

L:RegisterTranslations("enUS", function() return {
	["Modules"] = true,
	["Available Modules."] = true,
	["A module is missing its name field."] = true,
	["%s doesn't have a version field."] = true,
	["A module has been initialized twice for some reason."] = true,
	["Toggle whether to activate the %q Module."] = true,
	["%s has been disabled, since it is not compatible with the latest module API."] = true,
	["No modules running, please enable one or more from the configuration menu."] = true,
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------
local tablet = AceLibrary("Tablet-2.0")

local db = nil

local _G = getfenv(0)
local table_insert = table.insert

local moduleVersion = 1

local runningModules = {}

-----------------------------------------------------------------------
-- Local heap and utility functions
-----------------------------------------------------------------------

local new, del
do
	local cache = setmetatable({},{__mode='k'})
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select("#", ...) do
				table_insert(t, (select(i, ...)))
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				if type(v) == "table" then
					del(v)
				end
				t[k] = nil
			end
			cache[t] = true
		end
		return nil
	end
end

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local options = {
	type = "group",
	args = {
		tooltip = {
			type = "group",
			name = L["Modules"],
			desc = L["Available Modules."],
			args = {},
			order = 104,
		},
	},
}

-----------------------------------------------------------------------
-- Addon declaration
-----------------------------------------------------------------------
Expo = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "FuBarPlugin-2.0", "AceModuleCore-2.0")
local Expo = Expo

Expo:SetModuleMixins("AceEvent-2.0")
Expo.hasIcon = true

-----------------------------------------------------------------------
-- Module prototype
-----------------------------------------------------------------------

Expo.modulePrototype.core = Expo
Expo.modulePrototype.version = -1

local incompatible = {}
local registered = {}
function Expo.modulePrototype:OnInitialize()
	if type(self.name) ~= "string" then
		error(L["A module is missing its name field."])
	elseif type(self.version) ~= "number" then
		error(string.format(L["%s doesn't have a version field."],self.name))
	elseif registered[self.name] then
		error(L["A module has been initialized twice for some reason."])
	end

	if self.version == moduleVersion then
		self.core:OnRegister(self)
		registered[self.name] = true
	else
		incompatible[self] = true
		self.core:ModuleIncompatible(self)
	end
end

function Expo.modulePrototype:OnEnable(first)
	if incompatible[self] then return end

	if not runningModules[self.name] then

		if type(self.OnModuleEnable) == "function" then
			self:OnModuleEnable()
		end

		runningModules[self.name] = self

		self.core:TriggerEvent("Expo_ModuleEnabled",self.name)
	end
end

function Expo.modulePrototype:OnDisable()
	if runningModules[self.name] then
		if type(self.OnModuleDisable) == "function" then
			self:OnModuleDisable()
		end

		runningModules[self.name] = nil

		self.core:TriggerEvent("Expo_ModuleDisabled",self.name)
	end
end

function Expo.modulePrototype:new(...) return new(...) end
function Expo.modulePrototype:del(t) return del(t) end

-----------------------------------------------------------------------
-- Implicit module interface
-----------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local IExpo = AceOO.Interface{

	-- Optional methods:
	--
	-- -- Called when the module is enabled by the core. Any event registration
	-- -- should happen here.
	-- OnModuleEnable
	-- OnModuleDisable
	--
	-- -- Called when the module is initially registered.
	-- OnModuleRegister

	-- Must return the current module revision as a number.
	GetRevision = "function",

	-- Must return the module name as it should be displayed in the menu or tooltip, for
	-- example "Event Tracker".
	GetDisplayName = "function",

	-- Sent with the tablet category to use to fill with the module's data.
	GetData = "function",
	
	-- Returns the number of columns needed by the Module's tooltip display
	GetNumberOfColums = "function",
}

-----------------------------------------------------------------------
-- Addon initialization
-----------------------------------------------------------------------

function Expo:OnInitialize()
	local revision = tonumber(("$Revision: 32676 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.Expo_VERSION = self.version
	_G.Expo_REVISION = self.revision

	self.OnMenuRequest = options

	self:RegisterDB("ExpoDB")
	self:RegisterDefaults("account", {
		enabledModules = {
			["EventTracker"] = true,
			["EvalTracker"] = true,
		},
		activeModule = {
			["EventTracker"] = true,
			["EvalTracker"] = true,
		},
	})
end

function Expo:OnEnable(first)
	db = self.db.account

	-- Enable/disable all modules.
	for name, mod in self:IterateModules() do
		self:ToggleModuleActive(name, db.enabledModules[name])
	end

end

function Expo:OnDisable()
	-- Disable all modules.
	for name, mod in pairs(runningModules) do
		self:ToggleModuleActive(name, false)
	end
end

-----------------------------------------------------------------------
-- Module handling
-----------------------------------------------------------------------

function Expo:OnRegister(mod)
	--if not AceOO.inherits(mod, IExpo) then
		--error(string.format("The module %q does not inherit from the IExpo interface.", tostring(mod.name)))
	--end

	local moduleName = mod.name
	local displayName = mod:GetDisplayName()

	options.args.tooltip.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to activate the %q Module."], displayName),
		get = function()
			return self.db.account.enabledModules[moduleName]
		end,
		set = function(v)
			self.db.account.enabledModules[moduleName] = v
			self:ToggleModuleActive(moduleName, v)
			self:UpdateDisplay()
		end
	}

	if type(mod.OnModuleRegister) == "function" then
		mod:OnModuleRegister()
	end
end

function Expo:GetModuleVersion()
	return moduleVersion
end

function Expo:ModuleIncompatible(mod)
	self:ToggleModuleActive(mod, false)
	self:Print(string.format(L["%s has been disabled, since it is not compatible with the latest module API."], mod.name))
end

-----------------------------------------------------------------------
-- UI
-----------------------------------------------------------------------

function Expo:FormatValue(arg)
	local t = type(arg)
	if t == 'string' then
		local rearg = string.gsub(string.sub(arg,1,60), "|", "||")
		rearg = string.gsub(rearg, "\n", "\\n")
		arg = string.gsub(rearg, "\r", "\\r")
	elseif t == 'boolean' then
		arg = boolean and 'true' or 'false'
	elseif t ~= 'number' then
		return '<'..t..'>'
	end
	return '<'..t..'>: '..arg
end

function Expo:OnTooltipUpdate()
	-- sort through the list of modules, and put headers in for each one.
	-- the active module also gets polled for it's data.
	-- only one category should be open at a time, when one opens, all the rest are closed.
	local no = true
	tablet:AddCategory('hideBlankLine', true, 'text', 'Expo', 'justify', "CENTER",
		'font', GameTooltipHeaderText, 'isTitle', true)
	for k, v in pairs(runningModules) do
		no = nil
		local cat = tablet:AddCategory(
			'columns', v:GetNumberOfColumns(),
			'text', v:GetDisplayName(),
			'child_textR', 0.8,
			'child_textG', 0.8,
			'child_textB', 0.8,
			'child_text2R', 1,
			'child_text2G', 1,
			'child_text2B', 1,
			'child_size', 10,
			'child_size2', 12,
			'showWithoutChildren', true,
			'hideBlankLine', true,
			'checked',true,'hasCheck',true,
			'checkIcon',db.activeModule[k] and "Interface\\Buttons\\UI-PlusButton-Up" or "Interface\\Buttons\\UI-MinusButton-Up",
			'func',function() db.activeModule[k] = not db.activeModule[k] end
		)
		if db.activeModule[k] then
			v:GetData(cat)
		end
	end
	if no then
		tablet:AddCategory("columns", 1):AddLine("text", L["No modules running, please enable one or more from the configuration menu."])
	end
end
